from pathlib import Path
from typing import Optional, Set, Union, List

from crudhex.domain.config.SpecConfig import SpecConfig
from crudhex.domain.models import IdMeta, Relation, GenerationType, RelationType
from crudhex.domain.models import Entity, Field
from crudhex.domain.services.type_resolver import TypeResolver, get_type_resolver

from crudhex.domain.ports import spec_loader


def parse_spec_file(spec_path: Path) -> List[Entity]:
    spec_data = spec_loader.load_spec_config(spec_path)
    type_resolver = get_type_resolver()

    return parse_spec_data(spec_data, type_resolver)


def parse_spec_data(spec_data: dict, type_resolver: Optional[TypeResolver] = None) -> List[Entity]:
    entities = []
    # noinspection PyTypeChecker
    generated_types = spec_data.keys()  # type: set
    for name, spec in spec_data.items():
        entities.append(parse_entity_data(name, spec, generated_types, type_resolver))

    return entities


def parse_entity_data(name: str, data: dict, generated_types: Set[str],
                      type_resolver: Optional[TypeResolver]) -> Entity:
    entity = Entity(name)
    entity.meta = parse_entity_meta(data)

    fields_data = data.copy()
    if SpecConfig.META in fields_data: fields_data.pop(SpecConfig.META)

    for name, field_data in fields_data.items():
        entity.fields.append(parse_field(name, field_data, generated_types, type_resolver))

    return entity


def parse_entity_meta(data: dict) -> Optional[Entity.Meta]:
    meta = data.get(SpecConfig.META)
    if not meta or SpecConfig.TABLE not in meta: return None

    return Entity.Meta(meta[SpecConfig.TABLE])


def parse_field(name: str, data: Union[str, dict], generated_types: Set[str],
                type_resolver: Optional[TypeResolver]) -> Field:
    if type(data) is str:
        class_type = data if not type_resolver else type_resolver.resolve_type(data)
        return Field(name, class_type, class_type in generated_types)

    if SpecConfig.TYPE not in data: raise RuntimeError(f'Missing required key {SpecConfig.TYPE}')

    class_type = (data[SpecConfig.TYPE]
                  if not type_resolver
                  else type_resolver.resolve_type(data[SpecConfig.TYPE]))

    field = Field(name, class_type, class_type in generated_types)
    field.column = data.get(SpecConfig.COLUMN)
    field.id_meta = parse_field_id_meta(data)
    field.relation = parse_field_relation(data)

    # TODO: Configurable in dsl file
    if field.has_relation() and field.relation.type.has_multiple():
        field.type.collection_type = TypeResolver.DEFAULT_COLLECTION

    return field


def parse_field_id_meta(data: dict) -> Optional[IdMeta]:
    id_data = data.get(SpecConfig.ID)
    if not id_data: return None

    if type(id_data) is str:
        return IdMeta(GenerationType(id_data))

    id_meta = IdMeta(GenerationType(id_data[SpecConfig.TYPE]))
    id_meta.sequence = id_data.get(SpecConfig.SEQUENCE)

    return id_meta


def parse_field_relation(data: dict) -> Optional[Relation]:
    relation_data = data.get(SpecConfig.RELATION)
    if not relation_data: return None

    relation = Relation(RelationType(relation_data[SpecConfig.TYPE]))
    relation.join_table = relation_data.get(SpecConfig.JOIN_TABLE)
    relation.join_column = relation_data.get(SpecConfig.JOIN_COLUMN)
    relation.inverse_join_column = relation_data.get(SpecConfig.INVERSE_JOIN_COLUMN)
    relation.mapped_by = relation_data.get(SpecConfig.MAPPED_BY)

    return relation



from pathlib import Path
from typing import Dict, List, Optional

from crudhex.adapters.infrastructure.template_writer.config import template_config
from crudhex.adapters.infrastructure.template_writer.config.template_config import get_mapper_file_path
from crudhex.adapters.infrastructure.template_writer.services.template_env import get_template_environment
from crudhex.domain.models.mapper import MapperType


_MAPPERS_MAP = {
    MapperType.NONE: None,
    MapperType.MAPSTRUCT: template_config.MAPSTRUCT_TEMPLATE,
    MapperType.MODELMAPPER: None
}


def create_mapper(dest: Path, class_type: str, package: str,
                  imports: List[str], mappings: Dict[str, str], mapper_type: MapperType):

    if not _has_mapper(mapper_type): return

    template_env = get_template_environment()
    template = _get_mapper_template(mapper_type)
    if not template: return  # TODO: Exception?

    mapper_template = template_env.get_template(get_mapper_file_path(template))
    mapper_code = mapper_template.render({
        'package': package,
        'imports': '\n'.join(imports),
        'class_type': class_type,
        'mappings': mappings
    })

    with open(dest.resolve(), 'w+', encoding='utf-8') as f:
        f.write(mapper_code)


def _get_mapper_template(mapper_type: MapperType) -> Optional[str]:
    return _MAPPERS_MAP.get(mapper_type)


def _has_mapper(mapper_type: MapperType) -> bool:
    return mapper_type != MapperType.NONE



from pathlib import Path
from os import sep
from typing import Optional, Tuple

import typer
from rich.console import Console
from rich.theme import Theme
from rich.progress import Progress

from crudhex.domain.models.mapper import MapperType
from crudhex.domain.models.project_config import ConfigValidationError
from crudhex.domain.services import dsl_parser, config_context, db_adapter_generator, domain_generator, rest_generator


_SPEC_HELP = 'Spec file path to process'
_CONF_HELP = f'Project config file to know packages and code paths. Defaults to {config_context.DEFAULT_CONFIG}'
_FORCE_HELP = f'Override file outputs if exists'
_MAPPER_HELP = f'Mapper generation option. By default no mapper will be generated.' \
               '\n[WARN]: modelmapper generation not supported yet.'

app = typer.Typer()
out_console: Optional[Console] = None
err_console: Optional[Console] = None


def main():
    _setup()
    app()


@app.command()
def generate(
        spec_file: str = typer.Argument(..., help=_SPEC_HELP),
        project_config: str = typer.Option(None, '--config', '-c', help=_CONF_HELP),
        force_override: bool = typer.Option(False, '--force', '-f', help=_FORCE_HELP),
        mapper_type: MapperType = typer.Option(MapperType.NONE.value, '--mapper', '-m', help=_MAPPER_HELP)
):

    load_config(project_config)
    _validate_mapper(mapper_type)

    spec_path = Path(spec_file)
    if not spec_path.exists():
        err_console.print('Unable to find spec file: {}'.format(spec_path.resolve()))
        raise typer.Exit(code=1)

    if not force_override:
        out_console.print('Force flag not enabled, existing files won\'t be generated', style='warning')

    out_console.print('Processing spec...', end='\n\n', style='notify')
    with Progress(transient=True) as progress:
        parse_task = progress.add_task('Parsing config', total=100)
        entities = dsl_parser.parse_spec_file(spec_path)
        entities_map = {e.name: e for e in entities}
        progress.update(parse_task, advance=100)

        generate_task = progress.add_task('Generate classes', total=100)

        # Shared classes
        progress.console.rule('Shared classes', style='bright_yellow')
        if mapper_type != MapperType.NONE:
            out_path = db_adapter_generator.create_mapper_class(entities_map, mapper_type, force_override)
            _log_db_adapter_generation('DB mapper', out_path, progress)

        progress.console.print('')

        # Per entity classes
        for entity in entities:
            progress.console.rule(entity.name)
            out_path = domain_generator.create_model_class(entity, force_override)
            _log_domain_generation('Domain model', out_path, progress)

            out_path = domain_generator.create_command_class(entity, entities_map, force_override)
            _log_domain_generation('Domain command', out_path, progress)

            out_path = domain_generator.create_db_port_class(entity, force_override)
            _log_domain_generation('DB port', out_path, progress)

            out_path = domain_generator.create_use_case_port_class(entity, force_override)
            _log_domain_generation('Use case port', out_path, progress)

            out_path = domain_generator.create_use_case_class(entity, force_override)
            _log_domain_generation('Use case', out_path, progress)

            out_path = db_adapter_generator.create_entity_class(entity, force_override)
            _log_db_adapter_generation('Entity', out_path, progress)

            out_path = db_adapter_generator.create_repository_class(entity, force_override)
            _log_db_adapter_generation('Repository', out_path, progress)

            out_path = db_adapter_generator.create_adapter_class(entity, entities_map, force_override)
            _log_db_adapter_generation('DB adapter', out_path, progress)

            out_path = rest_generator.create_model_class(entity, force_override)
            _log_rest_adapter_generation('Controller', out_path, progress)

            progress.update(generate_task, advance=100/len(entities))
            progress.console.print('')

        progress.update(generate_task, advance=100)

    out_console.print('\n-- All classes generated --\n', style='finished')


def load_config(project_config: Optional[str]):
    """
    Loads config file. If config cannot be loaded or invalid will raise an exeption
    :param project_config: Config path. If `None` default path will be used
    :raises Exit
    """
    if not project_config:
        out_console.print('Loading from default config path...', style='info')

    try:
        config_context.load_config(Path(project_config) if project_config else None)
        config_context.get_config().validate()
    except ConfigValidationError as err:
        err_console.print('Errors in config file: {}'.format('\n- '.join(err.errors)))
        raise typer.Exit(code=1)


def _log_domain_generation(file_type: str, gen_output: Tuple[bool, Path], progress: Optional[Progress] = None):
    _log_generation(file_type, gen_output, 'bright_blue', progress)


def _log_db_adapter_generation(file_type: str, gen_output: Tuple[bool, Path], progress: Optional[Progress] = None):
    _log_generation(file_type, gen_output, 'bright_cyan', progress)


def _log_rest_adapter_generation(file_type: str, gen_output: Tuple[bool, Path], progress: Optional[Progress] = None):
    _log_generation(file_type, gen_output, 'bright_yellow', progress)


def _log_generation(file_type: str, gen_output: Tuple[bool, Path], style: str, progress: Optional[Progress] = None):
    console = out_console
    if progress:
        console = progress.console

    if gen_output[0]:
        console.print(f'{file_type}: [default]{gen_output[1].parent}{sep}[/default][{style}]{gen_output[1].name}[/{style}]', style=style)
    else:
        console.print(f'{file_type}: [yellow]Skipped[/yellow]', style=style)


def _validate_mapper(mapper_type: MapperType):
    if mapper_type == MapperType.MODELMAPPER:
        out_console.print(f"'{mapper_type.value}' mapper type not supported yet :(", style='critical')
        raise typer.Exit(code=1)


def _setup():
    theme = Theme({
        "notify": "cyan",
        "info": "dim cyan",
        "success": "green",
        "finished": "black on green",
        "warning": "yellow",
        "error": "red",
        "critical": "bold red"
    })

    global out_console
    out_console = Console(theme=theme)

    global err_console
    err_console = Console(stderr=True, style="red")



"""
Implements `ISharedClass` protocol interface
"""
from pathlib import Path
from typing import List, Dict

from crudhex.domain.models import Entity
from crudhex.domain.utils.class_type_utils import get_import, get_type_import, get_resolved_import
from crudhex.domain.utils import package_utils
from crudhex.domain.utils.file_utils import get_java_filename
from . import entity_generator
from ..config_context import get_config
from ..domain import model_generator
from crudhex.domain.ports import db_code_writer
from ...models.mapper import MapperType

_MAPPER_NAME = 'DbMapper'


def create_class(entities_map: Dict[str, Entity], mapper_type: MapperType, folder: Path) -> Path:
    if not folder.is_dir(): raise RuntimeError('Output path must be a folder ({})'.format(folder.resolve()))

    class_type = get_type_name()
    mapper_file = folder / get_java_filename(class_type)
    mappings = {entity_generator.get_type_name(e): model_generator.get_type_name(e) for e in entities_map.values()}

    db_code_writer.create_mapper(mapper_file, class_type, get_package(),
                                 _get_imports(entities_map), mappings, mapper_type)

    return mapper_file


def get_package() -> str:
    config = get_config()
    if config.db_mapper_class: return package_utils.get_package(config.db_mapper_class)

    return get_config().db_mapper_pkg


def get_type_name() -> str:
    config = get_config()
    if config.db_mapper_class: return package_utils.get_class_name(config.db_mapper_class)

    return _MAPPER_NAME


def get_filename() -> str:
    return get_java_filename(get_type_name())


def _get_imports(entities_map: Dict[str, Entity]) -> List[str]:
    imports = []
    for entity in entities_map.values():
        imports.extend([
            get_import(entity_generator.get_package(), entity_generator.get_type_name(entity)),
            get_import(model_generator.get_package(), model_generator.get_type_name(entity))
        ])

    return imports
